from rbfExcel.reader import ExcelReader
from rbfExcel.utils import BoolFormat, DateFormat, NumberFormat
from rbfExcel.writer import ExcelWriter


class ExcelLibrary(object):

    def __init__(self, date_format=DateFormat(), number_format=NumberFormat(), bool_format=BoolFormat()):
        """
        Init Excel Keyword with some default configuration.

        Excel Date Time format
        https://support.office.com/en-us/article/format-numbers-as-dates-or-times-418bd3fe-0577-47c8-8caa-b4d30c528309
        """
        self.date_format = date_format
        self.number_format = number_format
        self.bool_format = bool_format
        self.reader = None
        self.writer = None

    def open_excel(self, file_path):
        """
        *Opens the Excel file to read from the path provided in the file path parameter.*

        *Arguments*
        
        ``file_path`` : The Excel file name or path will be opened. If file name then opening file in current directory.
        
        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |

        """
        self.reader = ExcelReader(file_path, self.date_format, self.number_format, self.bool_format)

    def open_excel_to_write(self, file_path, new_path=None, override=False):
        """
        *Opens the Excel file to write from the path provided in the file name parameter.*
        
        In case `New Path` is given, new file will be created based on content of current file.

        *Arguments*
        
        ``file_path`` : The Excel file name or path will be opened. If file name then opening file in current directory.
        
        ``new_path`` : New path will be saved
        
        ``override`` : Default is False. If `True`, new file will be overriden if it exists.
        
        *Examples*

        | `Open Excel To Write`       |  ${CURDIR}/rbfExcelTest.xls  |

        """
        self.writer = ExcelWriter(file_path, new_path, override, self.date_format, self.number_format, self.bool_format)

    def get_sheet_names(self):
        """
        *Returns the names of all the worksheets in the current workbook.*

        *Examples*

        | `Open Excel`              |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Sheet Names`        |                                                    |

        """
        return self.reader.get_sheet_names()

    def get_number_of_sheets(self):
        """
        *Returns the number of worksheets in the current workbook.*

        *Examples*

        | `Open Excel`              |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Number Of Sheets`    |                                                    |

        """
        return self.reader.get_number_of_sheets()

    def get_column_count(self, sheet_name):
        """
        *Returns the specific number of columns of the sheet name specified.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the column count will be returned from
        
                
        *Examples*

        | `Open Excel`          |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Column Count`    |  Sheet1                                        |

        """
        return self.reader.get_column_count(sheet_name)

    def get_row_count(self, sheet_name):
        """
        *Returns the specific number of rows of the sheet name specified.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the column count will be returned from
        
        *Examples*

        | `Open Excel`          |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Row Count`       |  Sheet1                                        |

        """
        return self.reader.get_row_count(sheet_name)

    def get_column_values(self, sheet_name, column, include_empty_cells=True):
        """
        *Returns the specific column values of the sheet name specified.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the column count will be returned from
        
        ``column`` (int) : The column integer value is indicated to get values.
        
        ``include_empty_cells`` : Default is True. If `False` then only return cells with values. 
        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |   |
        | `Get Column Values`    |  Sheet1                                        | 0 |

        """
        return self.reader.get_column_values(sheet_name, column, include_empty_cells)

    def get_row_values(self, sheet_name, row, include_empty_cells=True):
        """
        *Returns the specific row values of the sheet name specified.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the column count will be returned from
        
        ``row`` (int) : The row integer value is indicated to get values.
        
        ``include_empty_cells`` : Default is True. If `False` then only return cells with values. 
        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |   |
        | `Get Row Values`       |  Sheet1                                        | 0 |

        """
        return self.reader.get_row_values(sheet_name, row, include_empty_cells)

    def get_sheet_values(self, sheet_name, include_empty_cells=True):
        """
        *Returns the values from the sheet name specified.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the column count will be returned from
        
        ``include_empty_cells`` : Default is True. If `False` then only return cells with values.
        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Sheet Values`     |  Sheet1 |

        """
        return self.reader.get_sheet_values(sheet_name, include_empty_cells)

    def get_workbook_values(self, include_empty_cells=True):
        """
        *Returns the values from each sheet of the current workbook.*

        *Arguments*
        
        ``include_empty_cells`` : Default is True. If `False` then only return cells with values.
        
        *Examples*
        
        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |
        | `Get Workbook Values`  |                                                    |

        """
        return self.reader.get_workbook_values(include_empty_cells)

    def read_cell_data_by_name(self, sheet_name, cell_name, data_type=None, use_format=True):
        """
        *Uses the cell name to return the data from that cell.*

        - `Data Type` indicates explicit data type to convert cell value to correct data type.
        - `Use Format` is False, then cell value will be raw data with correct data type.

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell value will be returned from.
        
        ``cell_name`` (String) : The selected cell name that the value will be returned from.
        
        ``data_type`` (String) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL
        
        ``use_format`` (boolean) : Default is True. Use format to convert data to string
        
      
        *Examples*

        | `Open Excel`                |  ${CURDIR}/rbfExcelTest.xls  |      |
        | `Read Cell Data By Name`    |  Sheet1                                        |  A2  |

        """
        return self.reader.read_cell_data_by_name(sheet_name, cell_name, data_type, use_format)

    
    def read_cell_data_by_row_and_column_header(self, sheet_name, row_header, column_header, data_type=None, use_format=True):
        """
        *Uses the row header name and column header name to return the data from the corresponding cell.*

        - `Data Type` indicates explicit data type to convert cell value to correct data type.
        - `Use Format` is False, then cell value will be raw data with correct data type.

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell value will be returned from.
        
        ``row_header`` (String) : Row Headers are usually defined in the first column of the excel sheet
		
		``column_header`` (String) : Column Headers are usually defined in the first row of the excel sheet
        
        ``data_type`` (String) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL
        
        ``use_format`` (boolean) : Default is True. Use format to convert data to string
        
      
        *Examples*

        | `Open Excel`                |  ${CURDIR}/rbfExcelTest.xls  |      |
        | ${value} | `Read Cell Data By Row And Column Header`    |  Sheet1 |  rowheadername  | columnheadername | 
        """
        return self.reader.read_cell_data_by_row_and_column_header(sheet_name, row_header, column_header, data_type, use_format)
		
    def read_cell_data(self, sheet_name, column, row, data_type=None, use_format=True):
        """
        *Uses the column and row to return the data from that cell.*

        - `Data Type` indicates explicit data type to convert cell value to correct data type.
        - `Use Format` is False, then cell value will be raw data with correct data type.
        
        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell value will be returned from.
        
        ``column`` (int) : The column integer value that the cell value will be returned from.
        
        ``row`` (int) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL
        
        ``data_type`` (String) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL
        
        ``use_format`` (boolean) : Default is True. Use format to convert data to string

        *Examples*

        | `Open Excel`        |  ${CURDIR}/rbfExcelTest.xls  |   |   |
        | `Read Cell Data`    |  Sheet1                                        | 0 | 0 |

        """
        return self.reader.read_cell_data(sheet_name, column, row, data_type, use_format)

    def check_cell_type(self, sheet_name, column, row, data_type):
        """
        *Checks the type of value that is within the cell of the sheet name selected.*

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell value will be returned from.
        
        ``column`` (int) : The column integer value that the cell value will be returned from.
        
        ``row`` (int) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL
        
        ``data_type`` (String) : Available options: DATE , TIME , DATE_TIME , TEXT , NUMBER , BOOL , EMPTY , ERROR
        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |   |   |       |
        | `Check Cell Type`      |  Sheet1                                        | 0 | 0 | DATE  |

        """
        return self.reader.check_cell_type(sheet_name, column, row, data_type)

    def write_to_cell_by_name(self, sheet_name, cell_name, value, data_type=None):
        """
        *Write data to cell by using the given sheet name and the given cell that defines by name.*

        If `Data Type` is not provided, `rbfExcel` will introspect data type from given `value` to define cell type

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell value will be returned from.
        
        ``cell_name`` (String) : The selected cell name that the value will be returned from.
        
        ``value`` String|number|datetime|boolean) : Raw value or string value then using DataType to decide data type to write
        
        ``data_type`` (String) : Available options: DATE , TIME , DATE_TIME , TEXT , NUMBER , BOOL
        
        *Examples*

        | `Open Excel`            |  ${CURDIR}/rbfExcelTest.xls  |                      |       |
        | `Write To Cell By Name` |  Sheet1                                        |  A1  |  34           |       |
        | `Write To Cell By Name` |  Sheet1                                        |  A2  |  2018-03-29   | DATE  |
        | `Write To Cell By Name` |  Sheet1                                        |  A3  |  YES          | BOOL  |

        """
        self.writer.write_to_cell_by_name(sheet_name, cell_name, value, data_type)

    def write_to_cell_by_row_and_column_header(self, sheet_name, row_header, column_header, value, data_type=None):
        """
        *Uses the row header name and column header name to write the data to the corresponding cell.*

        - `Data Type` indicates explicit data type to convert cell value to correct data type.
        - `Use Format` is False, then cell value will be raw data with correct data type.

        *Arguments*

        ``sheet_name`` (String) : The selected sheet that the cell value will be written to.

        ``row_header`` (String) : Row Headers are usually defined in the first column of the excel sheet

        ``column_header`` (String) : Column Headers are usually defined in the first row of the excel sheet

        ``value`` String|number|datetime|boolean) : Raw value or string value then using DataType to decide data type to write

        ``data_type`` (String) : Available options: TEXT , DATE , TIME , DATETIME , NUMBER , BOOL


        *Examples*

        | `Open Excel`                |  ${CURDIR}/rbfExcelTest.xls  |      |
        | ${value} | `Write To Cell By Row And Column Header`    |  Sheet1 |  rowheadername  | columnheadername | value |

        """
        print("in base class")
        self.writer.write_to_cell_by_row_and_column_header(sheet_name, row_header, column_header, value, data_type)
	
    def write_to_cell(self, sheet_name, column, row, value, data_type=None):
        """
        *Write data to cell by using the given sheet name and the given cell that defines by column and row.*

        If `Data Type` is not provided, `rbfExcel` will introspect data type from given `value` to define cell type

        *Arguments*
        
        ``sheet_name`` (String) : The selected sheet that the cell will be modified from.
        
        ``column`` (int) : The column integer value that the cell value will be returned from.
        
        ``row`` (int) : The row integer value that will be used to modify the cell.
        
        ``value`` (String|number|datetime|boolean) : Raw value or string value then using DataType to decide data type to write
        
        ``data_type`` (String) : Available options: DATE , TIME , DATE_TIME , TEXT , NUMBER , BOOL
        
        *Examples*

        | `Open Excel`            |  ${CURDIR}/rbfExcelTest.xls  |     |     |              |       |
        | `Write To Cell`         |  Sheet1                                        |  0  |  0  |  34          |       |
        | `Write To Cell`         |  Sheet1                                        |  1  |  1  |  2018-03-29  | DATE  |
        | `Write To Cell`         |  Sheet1                                        |  2  |  2  |  YES         | BOOL  |

        """
        self.writer.write_to_cell(sheet_name, column, row, value, data_type)


    def save_excel(self):
        """
        *Saves the Excel file that was opened to write before.*

        *Examples*
        
        | `Open Excel To Write`   |  ${CURDIR}/rbfExcelTest.xls  |                  |
        | `Write To Cell`         |  Sheet1                                        |  0  |  0  |  34  |
        | `Save Excel`            |                                                    |                  |

        """
        self.writer.save_excel()

    def create_sheet(self, sheet_name):
        """
        *Creates and appends new Excel worksheet using the new sheet name to the current workbook.*

        *Arguments*
        
        ``sheet_name`` (String) : The name of the new sheet added to the workbook.

        
        *Examples*

        | `Open Excel`           |  ${CURDIR}/rbfExcelTest.xls  |
        | `Create Sheet`         |  NewSheet                                          |

        """
        self.writer.create_sheet(sheet_name)



#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

import logging
import os
import six
import socket
import subprocess
import time
import warnings

log = logging.getLogger(__name__)


try:
    BrokenPipeError
except NameError:
    class BrokenPipeError(Exception):
        pass


"""
    Python 3+ note: unicode strings should be used when communicating with the Emulator methods.
    utf-8 is used internally when reading from or writing to the 3270 emulator (this includes
    reading lines, constructing data to write, reading statuses).
"""


class CommandError(Exception):
    pass


class TerminatedError(Exception):
    pass


class WaitError(Exception):
    pass


class KeyboardStateError(Exception):
    pass


class FieldTruncateError(Exception):
    pass


class Command(object):
    """
        Represents a x3270 script command
    """
    def __init__(self, app, cmdstr):
        if isinstance(cmdstr, six.text_type):
            warnings.warn('Commands should be byte strings', stacklevel=3)
            cmdstr = cmdstr.encode("utf-8")
        self.app = app
        self.cmdstr = cmdstr
        self.status_line = None
        self.data = []

    def execute(self):
        self.app.write(self.cmdstr + b'\n')

        # x3270 puts data lines (if any) on stdout prefixed with 'data: '
        # followed by two more lines without the prefix.
        # 1: status of the emulator
        # 2: 'ok' or 'error' indicating whether the command succeeded or failed
        while True:
            line = self.app.readline()
            # log.debug('stdout line: %s', line.rstrip())       # commented line to reduce log size
            if not line.startswith('data:'.encode("ascii")):
                # ok, we are at the status line
                self.status_line = line.rstrip()
                result = self.app.readline().rstrip()
                # log.debug('result line: %s', result)          # commented line to reduce log size
                return self.handle_result(result.decode("utf-8"))

            # remove the 'data: ' prefix and trailing newline char(s) and store
            self.data.append(line[6:].rstrip('\n\r'.encode("utf-8")))

    def handle_result(self, result):
        # should receive 'ok' for almost everything, but Quit returns a '' for
        # some reason
        if result == '' and self.cmdstr == b'Quit':
            return
        if result == 'ok':
            return
        if result != 'error':
            raise ValueError('expected "ok" or "error" result, but received: {0}'.format(result))

        msg = b'[no error message]'
        if self.data:
            msg = ''.encode("utf-8").join(self.data).rstrip()
        raise CommandError(msg.decode("utf-8"))


class Status(object):
    """
        Represents a status line as returned by x3270 following a command
    """
    def __init__(self, status_line):
        if not status_line:
            status_line = (' ' * 12).encode("utf-8")
        parts = status_line.split(' '.encode("utf-8"))
        self.as_string = status_line.rstrip().decode("utf-8")
        self.keyboard = parts[0] or None
        self.screen_format = parts[1] or None
        self.field_protection = parts[2] or None
        self.connection_state = parts[3] or None
        self.emulator_mode = parts[4] or None
        self.model_number = parts[5] or None
        self.row_number = parts[6] or None
        self.col_number = parts[7] or None
        self.cursor_row = parts[8] or None
        self.cursor_col = parts[9] or None
        self.window_id = parts[10] or None
        self.exec_time = parts[11] or None

    def __str__(self):
        return 'STATUS: {0}'.format(self.as_string)


class ExecutableAppLinux(object):
    executable = None
    args = [
        '-xrm', 's3270.unlockDelay: False'
    ]

    def __init__(self):
        self.sp = None
        self.spawn_app()

    def spawn_app(self):
        args = [self.executable] + self.args
        self.sp = subprocess.Popen(
            args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def connect(self, host):
        """ this is a no-op for all but wc3270 """
        return False

    def close(self):
        pass

    def write(self, data):
        self.sp.stdin.write(data)
        self.sp.stdin.flush()

    def readline(self):
        return self.sp.stdout.readline()


class ExecutableAppWin(object):
    executable = None
    args = [
        '-xrm', 'wc3270.unlockDelay: False'
    ]
    script_port = 17938

    def __init__(self):
        self.sp = None
        self.socket_fh = None

    def connect(self, host):
        self.spawn_app(host)
        self.make_socket()
        return True

    def close(self):
        # failing to close the socket ourselves will result in a ResourceWarning
        self.socket.close()
        
    def spawn_app(self, host):
        args = ['start', '/wait', self.executable] + self.args
        args.extend(['-scriptport', str(self.script_port), host])
        self.sp = subprocess.Popen(
            args,
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    def make_socket(self):
        self.socket = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        count = 0
        while count < 15:
            try:
                sock.connect(('127.0.0.1', self.script_port))
                break
            except socket.error as e:
                if 'actively refused it' not in str(e):
                    raise
                time.sleep(1)
                count += 1
        # open a file handle for the socket that can both read and write, using bytestrings
        self.socket_fh = sock.makefile(mode='rwb')

    def write(self, data):
        if self.socket_fh is None:
            raise NotConnectedException
        self.socket_fh.write(data)
        self.socket_fh.flush()

    def readline(self):
        if self.socket_fh is None:
            raise NotConnectedException
        return self.socket_fh.readline()


class x3270App(ExecutableAppLinux):
    executable = 'x3270'
    # Per Paul Mattes, in the first days of x3270, there were servers that
    # would unlock the keyboard before they had processed the command. To
    # work around that, when AID commands are sent, there is a 350ms delay
    # before the command returns. This arg turns that feature off for
    # performance reasons.
    args = ['-xrm', 'x3270.unlockDelay: False', '-xrm', 'x3270.model: 2', '-script']


class s3270App(ExecutableAppLinux):
    executable = 's3270'
    # see notes for args in x3270App
    args = ['-xrm', 's3270.unlockDelay: False']


class NotConnectedException(Exception):
    pass


class wc3270App(ExecutableAppWin):
    executable = 'wc3270'
    # see notes for args in x3270App
    args = ['-xrm', 'wc3270.unlockDelay: False', '-xrm', 'wc3270.model: 2']


class ws3270App(ExecutableAppWin):
    executable = 'ws3270'
    # see notes for args in x3270App
    args = ['-xrm', 'ws3270.unlockDelay: False', ]


class EmulatorBase(object):
    def __init__(self, _sp=None):
        raise Exception("EmulatorBase has been replaced by Emulator.  See readme.rst.")


class Emulator(object):
    """
        Represents an x/s3270 emulator subprocess and provides an API for interacting
        with it.
    """

    def __init__(self, visible=False, timeout=30, app=None, _sp=None):
        """
            Create an emulator instance

            `visible` controls which executable will be used.
            `timeout` controls the timeout paramater to any Wait() command sent
                to x3270.
            `_sp` is normally not used but can be set to a mock object
                during testing.
        """
        self.app = app or self.create_app(visible)
        self.is_terminated = False
        self.status = Status(None)
        self.timeout = timeout
        self.last_host = None

    def __del__(self):
        """
            Since an emulator creates a process (and sometimes a socket handle), it is good practice
            to clean these up when done. Note, not terminating at this point will usually have no
            ill effect - only Python 3+ on Windows had problems in this regard.
        """
        self.terminate()

    def create_app(self, visible):
        if os.name == 'nt':
            if visible:
                return wc3270App()
            return ws3270App()
        if visible:
            return x3270App()
        return s3270App()

    def exec_command(self, cmdstr):
        """
            Execute an x3270 command

            `cmdstr` gets sent directly to the x3270 subprocess on it's stdin.
        """
        if self.is_terminated:
            raise TerminatedError('this TerminalClient instance has been terminated')

        # log.debug('sending command: %s', cmdstr)             # commented line to reduce log size
        c = Command(self.app, cmdstr)
        # start = time.time()                                  # unnecessary variable if the log is commented out.
        c.execute()
        # elapsed = time.time() - start                        # unnecessary variable if the log is commented out.
        # log.debug('elapsed execution: {0}'.format(elapsed))  # commented line to reduce log size
        self.status = Status(c.status_line)

        return c

    def terminate(self):
        """
            terminates the underlying x3270 subprocess. Once called, this
            Emulator instance must no longer be used.
        """
        if not self.is_terminated:
            log.debug('terminal client terminated')
            try:
                self.exec_command(b'Quit')
            except BrokenPipeError:  # noqa
                # x3270 was terminated, since we are just quitting anyway, ignore it.
                pass
            except socket.error as e:
                if 'was forcibly closed' not in str(e):
                    raise
                # this can happen because wc3270 closes the socket before
                # the read() can happen, causing a socket error

            self.app.close()

            self.is_terminated = True

    def is_connected(self):
        """
            Return bool indicating connection state
        """
        # need to wrap in try/except b/c of wc3270's socket connection dynamics
        try:
            # this is basically a no-op, but it results in the the current status
            # getting updated
            self.exec_command(b'ignore')

            # connected status is like 'C(192.168.1.1)', disconnected is 'N'
            return self.status.connection_state.startswith(b'C(')
        except NotConnectedException:
            return False

    def connect(self, host):
        """
            Connect to a host
        """
        if not self.app.connect(host):
            command = 'Connect({0})'.format(host).encode("utf-8")
            self.exec_command(command)
        self.last_host = host

    def reconnect(self):
        """
            Disconnect from the host and re-connect to the same host
        """
        self.exec_command(b'Disconnect')
        self.connect(self.last_host)

    def wait_for_field(self):
        """
            Wait until the screen is ready, the cursor has been positioned
            on a modifiable field, and the keyboard is unlocked.

            Sometimes the server will "unlock" the keyboard but the screen will
            not yet be ready.  In that case, an attempt to read or write to the
            screen will result in a 'E' keyboard status because we tried to
            read from a screen that is not yet ready.

            Using this method tells the client to wait until a field is
            detected and the cursor has been positioned on it.
        """
        self.exec_command('Wait({0}, InputField)'.format(self.timeout).encode("utf-8"))
        if self.status.keyboard != b'U':
            raise KeyboardStateError('keyboard not unlocked, state was: {0}'.format(
                self.status.keyboard.decode("utf-8")))

    def move_to(self, ypos, xpos):
        """
            move the cursor to the given co-ordinates.  Co-ordinates are 1
            based, as listed in the status area of the terminal.
        """
        # the screen's co-ordinates are 1 based, but the command is 0 based
        xpos -= 1
        ypos -= 1
        self.exec_command('MoveCursor({0}, {1})'.format(ypos, xpos).encode("utf-8"))

    def send_string(self, tosend, ypos=None, xpos=None):
        """
            Send a string to the screen at the current cursor location or at
            screen co-ordinates `ypos`/`xpos` if they are both given.

            Co-ordinates are 1 based, as listed in the status area of the
            terminal.
        """
        if xpos is not None and ypos is not None:
            self.move_to(ypos, xpos)

        # escape double quotes in the data to send
        tosend = tosend.decode("utf-8").replace('"', '\"')

        self.exec_command('String("{0}")'.format(tosend).encode("utf-8"))

    def send_enter(self):
        self.exec_command(b'Enter')

    def send_pf3(self):
        self.exec_command(b'PF(3)')

    def send_pf4(self):
        self.exec_command(b'PF(4)')

    def send_pf5(self):
        self.exec_command(b'PF(5)')

    def send_pf6(self):
        self.exec_command(b'PF(6)')

    def send_pf7(self):
        self.exec_command(b'PF(7)')

    def send_pf8(self):
        self.exec_command(b'PF(8)')

    def string_get(self, ypos, xpos, length):
        """
            Get a string of `length` at screen co-ordinates `ypos`/`xpos`

            Co-ordinates are 1 based, as listed in the status area of the
            terminal.
        """
        # the screen's co-ordinates are 1 based, but the command is 0 based
        xpos -= 1
        ypos -= 1
        cmd = self.exec_command('ascii({0},{1},{2})'.format(ypos, xpos, length).encode("utf-8"))
        # this usage of utf-8 should only return a single line of data
        assert len(cmd.data) == 1, cmd.data
        #return cmd.data[0].decode("utf-8")
        new_cmd_data=cmd.data[0].decode("cp1251").encode("utf-8")
        return new_cmd_data.decode("utf-8")


    def string_found(self, ypos, xpos, string):
        """
            Return True if `string` is found at screen co-ordinates
            `ypos`/`xpos`, False otherwise.

            Co-ordinates are 1 based, as listed in the status area of the
            terminal.
        """
        found = self.string_get(ypos, xpos, len(string))
        log.debug('string_found() saw "{0}"'.format(found))
        return found == string

    def delete_field(self):
        """
            Delete contents in field at current cursor location and positions
            cursor at beginning of field.
        """
        self.exec_command(b'DeleteField')

    def fill_field(self, ypos, xpos, tosend, length):
        """
            clears the field at the position given and inserts the string
            `tosend`

            tosend: the string to insert
            length: the length of the field

            Co-ordinates are 1 based, as listed in the status area of the
            terminal.

            raises: FieldTruncateError if `tosend` is longer than
                `length`.
        """
        if length - len(tosend) < 0:
            raise FieldTruncateError('length limit %d, but got "%s"' % (length, tosend))
        if xpos is not None and ypos is not None:
            self.move_to(ypos, xpos)
        self.delete_field()
        self.send_string(tosend)

    def save_screen(self, file_path):
        self.exec_command('PrintText(html,file,{0})'.format(file_path).encode("utf-8"))


# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.errors import ElementNotFound
from rbfWEB.utils import is_noney


class FormElementKeywords(LibraryComponent):

    @keyword
    def submit_form(self, locator=None):
        """Submits a form identified by ``locator``.

        If ``locator`` is not given, first form on the page is submitted.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Submitting form '%s'." % locator)
        if is_noney(locator):
            locator = 'tag:form'
        element = self.find_element(locator, tag='form')
        element.submit()

    @keyword
    def checkbox_should_be_selected(self, locator):
        """Verifies checkbox ``locator`` is selected/checked.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Verifying checkbox '%s' is selected." % locator)
        element = self._get_checkbox(locator)
        if not element.is_selected():
            raise AssertionError("Checkbox '%s' should have been selected "
                                 "but was not." % locator)

    @keyword
    def checkbox_should_not_be_selected(self, locator):
        """Verifies checkbox ``locator`` is not selected/checked.

        See the `Locating elements` section for details about the locator
        syntax..
        """
        self.info("Verifying checkbox '%s' is not selected." % locator)
        element = self._get_checkbox(locator)
        if element.is_selected():
            raise AssertionError("Checkbox '%s' should not have been "
                                 "selected." % locator)

    @keyword
    def page_should_contain_checkbox(self, locator, message=None, loglevel='TRACE'):
        """Verifies checkbox ``locator`` is found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_contains(locator, 'checkbox', message, loglevel)

    @keyword
    def page_should_not_contain_checkbox(self, locator, message=None, loglevel='TRACE'):
        """Verifies checkbox ``locator`` is not found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_not_contains(locator, 'checkbox', message, loglevel)

    @keyword
    def select_checkbox(self, locator):
        """Selects checkbox identified by ``locator``.

        Does nothing if checkbox is already selected.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Selecting checkbox '%s'." % locator)
        element = self._get_checkbox(locator)
        if not element.is_selected():
            element.click()

    @keyword
    def unselect_checkbox(self, locator):
        """Removes selection of checkbox identified by ``locator``.

        Does nothing if the checkbox is not selected.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Unselecting checkbox '%s'." % locator)
        element = self._get_checkbox(locator)
        if element.is_selected():
            element.click()

    @keyword
    def page_should_contain_radio_button(self, locator, message=None, loglevel='TRACE'):
        """Verifies radio button ``locator`` is found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, radio buttons are
        searched using ``id``, ``name`` and ``value``.
        """
        self.assert_page_contains(locator, 'radio button', message, loglevel)

    @keyword
    def page_should_not_contain_radio_button(self, locator, message=None, loglevel='TRACE'):
        """Verifies radio button ``locator`` is not found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, radio buttons are
        searched using ``id``, ``name`` and ``value``.
        """
        self.assert_page_not_contains(locator, 'radio button', message,
                                      loglevel)

    @keyword
    def radio_button_should_be_set_to(self, group_name, value):
        """Verifies radio button group ``group_name`` is set to ``value``.

        ``group_name`` is the ``name`` of the radio button group.
        """
        self.info("Verifying radio button '%s' has selection '%s'."
                  % (group_name, value))
        elements = self._get_radio_buttons(group_name)
        actual_value = self._get_value_from_radio_buttons(elements)
        if actual_value is None or actual_value != value:
            raise AssertionError("Selection of radio button '%s' should have "
                                 "been '%s' but was '%s'."
                                 % (group_name, value, actual_value))

    @keyword
    def radio_button_should_not_be_selected(self, group_name):
        """Verifies radio button group ``group_name`` has no selection.

        ``group_name`` is the ``name`` of the radio button group.
        """
        self.info("Verifying radio button '%s' has no selection." % group_name)
        elements = self._get_radio_buttons(group_name)
        actual_value = self._get_value_from_radio_buttons(elements)
        if actual_value is not None:
            raise AssertionError("Radio button group '%s' should not have "
                                 "had selection, but '%s' was selected."
                                 % (group_name, actual_value))

    @keyword
    def select_radio_button(self, group_name, value):
        """Sets radio button group ``group_name`` to ``value``.

        The radio button to be selected is located by two arguments:
        - ``group_name`` is the name of the radio button group.
        - ``value`` is the ``id`` or ``value`` attribute of the actual
          radio button.

        Examples:
        | `Select Radio Button` | size    | XL    |
        | `Select Radio Button` | contact | email |
        """
        self.info("Selecting '%s' from radio button '%s'."
                  % (value, group_name))
        element = self._get_radio_button_with_value(group_name, value)
        if not element.is_selected():
            element.click()

    @keyword
    def choose_file(self, locator, file_path):
        """Inputs the ``file_path`` into file input field ``locator``.

        This keyword is most often used to input files into upload forms.
        The file specified with ``file_path`` must be available on machine
        where tests are executed.

        Example:
        | `Choose File` | my_upload_field | ${CURDIR}/trades.csv |
        """
        if not os.path.isfile(file_path):
            raise ValueError("File '%s' does not exist on the local file "
                             "system." % file_path)
        self.find_element(locator).send_keys(file_path)

    @keyword
    def input_password(self, locator, password):
        """Types the given password into text field identified by ``locator``.

        See the `Locating elements` section for details about the locator
        syntax.

        Difference compared to `Input Text` is that this keyword does not
        log the given password on the INFO level. Notice that if you use
        the keyword like

        | Input Password | password_field | password |

        the password is shown as a normal keyword argument. A way to avoid
        that is using variables like

        | Input Password | password_field | ${PASSWORD} |

        Notice also that rbfWEB logs all the communication with
        browser drivers using the DEBUG level, and the actual password can
        be seen there. Additionally Robot Framework logs all arguments using
        the TRACE level. Tests must thus not be executed using level below
        INFO if password should not be logged in any format.
        """
        self.info("Typing password into text field '%s'." % locator)
        self._input_text_into_text_field(locator, password)

    @keyword
    def input_text(self, locator, text):
        """Types the given ``text`` into text field identified by ``locator``.

        Use `Input Password` if you do not want the given ``text`` to be
        logged.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.info("Typing text '%s' into text field '%s'." % (text, locator))
        self._input_text_into_text_field(locator, text)

    @keyword
    def page_should_contain_textfield(self, locator, message=None, loglevel='TRACE'):
        """Verifies text field ``locator`` is found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_contains(locator, 'text field', message, loglevel)

    @keyword
    def page_should_not_contain_textfield(self, locator, message=None, loglevel='TRACE'):
        """Verifies text field ``locator`` is not found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        self.assert_page_not_contains(locator, 'text field', message, loglevel)

    @keyword
    def textfield_should_contain(self, locator, expected, message=None):
        """Verifies text field ``locator`` contains text ``expected``.

        ``message`` can be used to override the default error message.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        actual = self._get_value(locator, 'text field')
        if expected not in actual:
            if is_noney(message):
                message = "Text field '%s' should have contained text '%s' "\
                          "but it contained '%s'." % (locator, expected, actual)
            raise AssertionError(message)
        self.info("Text field '%s' contains text '%s'." % (locator, expected))

    @keyword
    def textfield_value_should_be(self, locator, expected, message=None):
        """Verifies text field ``locator`` has exactly text ``expected``.

        ``message`` can be used to override default error message.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        actual = self._get_value(locator, 'text field')
        if actual != expected:
            if is_noney(message):
                message = "Value of text field '%s' should have been '%s' "\
                          "but was '%s'." % (locator, expected, actual)
            raise AssertionError(message)
        self.info("Content of text field '%s' is '%s'." % (locator, expected))

    @keyword
    def textarea_should_contain(self, locator, expected, message=None):
        """Verifies text area ``locator`` contains text ``expected``.

        ``message`` can be used to override default error message.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        actual = self._get_value(locator, 'text area')
        if expected not in actual:
            if is_noney(message):
                message = "Text area '%s' should have contained text '%s' " \
                          "but it had '%s'." % (locator, expected, actual)
            raise AssertionError(message)
        self.info("Text area '%s' contains text '%s'." % (locator, expected))

    @keyword
    def textarea_value_should_be(self, locator, expected, message=None):
        """Verifies text area ``locator`` has exactly text ``expected``.

        ``message`` can be used to override default error message.

        See the `Locating elements` section for details about the locator
        syntax.
        """
        actual = self._get_value(locator, 'text area')
        if expected != actual:
            if is_noney(message):
                message = "Text area '%s' should have had text '%s' " \
                          "but it had '%s'." % (locator, expected, actual)
            raise AssertionError(message)
        self.info("Content of text area '%s' is '%s'." % (locator, expected))

    @keyword
    def page_should_contain_button(self, locator, message=None, loglevel='TRACE'):
        """Verifies button ``locator`` is found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, buttons are
        searched using ``id``, ``name`` and ``value``.
        """
        try:
            self.assert_page_contains(locator, 'input', message, loglevel)
        except AssertionError:
            self.assert_page_contains(locator, 'button', message, loglevel)

    @keyword
    def page_should_not_contain_button(self, locator, message=None, loglevel='TRACE'):
        """Verifies button ``locator`` is not found from current page.

        See `Page Should Contain Element` for explanation about ``message``
        and ``loglevel`` arguments.

        See the `Locating elements` section for details about the locator
        syntax. When using the default locator strategy, buttons are
        searched using ``id``, ``name`` and ``value``.
        """
        self.assert_page_not_contains(locator, 'button', message, loglevel)
        self.assert_page_not_contains(locator, 'input', message, loglevel)

    def _get_value(self, locator, tag):
        return self.find_element(locator, tag).get_attribute('value')

    def _get_checkbox(self, locator):
        return self.find_element(locator, tag='checkbox')

    def _get_radio_buttons(self, group_name):
        xpath = "xpath://input[@type='radio' and @name='%s']" % group_name
        self.debug('Radio group locator: ' + xpath)
        elements = self.find_elements(xpath)
        if not elements:
            raise ElementNotFound("No radio button with name '%s' found."
                                  % group_name)
        return elements

    def _get_radio_button_with_value(self, group_name, value):
        xpath = "xpath://input[@type='radio' and @name='%s' and " \
                "(@value='%s' or @id='%s')]" % (group_name, value, value)
        self.debug('Radio group locator: ' + xpath)
        try:
            return self.find_element(xpath)
        except ElementNotFound:
            raise ElementNotFound("No radio button with name '%s' and "
                                  "value '%s' found." % (group_name, value))

    def _get_value_from_radio_buttons(self, elements):
        for element in elements:
            if element.is_selected():
                return element.get_attribute('value')
        return None

    def _input_text_into_text_field(self, locator, text):
        element = self.find_element(locator)
        element.clear()
        element.send_keys(text)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn

from rbfWEB.base import DynamicCore
from rbfWEB.errors import NoOpenBrowser
from rbfWEB.keywords import (AlertKeywords,
                                      BrowserManagementKeywords,
                                      CookieKeywords,
                                      ElementKeywords,
                                      FormElementKeywords,
                                      FrameKeywords,
                                      JavaScriptKeywords,
                                      RunOnFailureKeywords,
                                      ScreenshotKeywords,
                                      SelectElementKeywords,
                                      TableElementKeywords,
                                      WaitingKeywords,
                                      WebDriverCache,
                                      WindowKeywords,
                                      NettbankKeywords)
from rbfWEB.locators import ElementFinder
from rbfWEB.utils import Deprecated, LibraryListener, timestr_to_secs

from rbfWEB.version import __version__



class rbfWEB(DynamicCore):
    """rbfWEB is a web testing library for Robot Framework.

    This document explains how to use keywords provided by rbfWEB.
    

    rbfWEB uses the Selenium WebDriver modules internally to
    control a web browser. See http://seleniumhq.org for more information
    about Selenium in general.

    == Table of contents ==

    - `Locating elements`
    - `Timeouts, waits and delays`
    - `Run-on-failure functionality`
    - `Boolean arguments`
    - `Thread support`
    - `Importing`
    - `Shortcuts`
    - `Keywords`

    = Locating elements =

    All keywords in rbfWEB that need to interact with an element
    on a web page take an argument typically named ``locator`` that specifies
    how to find the element. Most often the locator is given as a string
    using the locator syntax described below, but `using WebElements` is
    possible too.

    == Locator syntax ==

    rbfWEB supports finding elements based on different strategies
    such as the element id, XPath expressions, or CSS selectors. The strategy
    can either be explicitly specified with a prefix or the strategy can be
    implicit.

    === Default locator strategy ===

    By default locators are considered to use the keyword specific default
    locator strategy. All keywords support finding elements based on ``id``
    and ``name`` attributes, but some keywords support additional attributes
    or other values that make sense in their context. For example, `Click
    Link` supports the ``href`` attribute and the link text and addition
    to the normal ``id`` and ``name``.

    Examples:

    | `Click Element` | example | # Match based on ``id`` or ``name``.            |
    | `Click Link`    | example | # Match also based on link text and ``href``.   |
    | `Click Button`  | example | # Match based on ``id``, ``name`` or ``value``. |

    If a locator accidentally starts with a prefix recognized as `explicit
    locator strategy` or `implicit XPath strategy`, it is possible to use
    the explicit ``default`` prefix to enable the default strategy.

    Examples:

    | `Click Element` | name:foo         | # Find element with name ``foo``.               |
    | `Click Element` | default:name:foo | # Use default strategy with value ``name:foo``. |
    | `Click Element` | //foo            | # Find element using XPath ``//foo``.           |
    | `Click Element` | default: //foo   | # Use default strategy with value ``//foo``.    |

    === Explicit locator strategy ===

    The explicit locator strategy is specified with a prefix using either
    syntax ``strategy:value`` or ``strategy=value``. The former syntax
    is preferred, because the latter is identical to Robot Framework's
    [http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#named-argument-syntax|
    named argument syntax] and that can cause problems. Spaces around
    the separator are ignored, so ``id:foo``, ``id: foo`` and ``id : foo``
    are all equivalent.

    Locator strategies that are supported by default are listed in the table
    below. In addition to them, it is possible to register `custom locators`.

    | = Strategy = |          = Match based on =         |         = Example =            |
    | id           | Element ``id``.                     | ``id:example``                 |
    | name         | ``name`` attribute.                 | ``name:example``               |
    | identifier   | Either ``id`` or ``name``.          | ``identifier:example``         |
    | class        | Element ``class``.                  | ``class:example``              |
    | tag          | Tag name.                           | ``tag:div``                    |
    | xpath        | XPath expression.                   | ``xpath://div[@id="example"]`` |
    | css          | CSS selector.                       | ``css:div#example``            |
    | dom          | DOM expression.                     | ``dom:document.images[5]``     |
    | link         | Exact text a link has.              | ``link:The example``           |
    | partial link | Partial link text.                  | ``partial link:he ex``         |
    | sizzle       | Sizzle selector provided by jQuery. | ``sizzle:div.example``         |
    | jquery       | Same as the above.                  | ``jquery:div.example``         |
    | default      | Keyword specific default behavior.  | ``default:example``            |

    See the `Default locator strategy` section below for more information
    about how the default strategy works. Using the explicit ``default``
    prefix is only necessary if the locator value itself accidentally
    matches some of the explicit strategies.

    Different locator strategies have different pros and cons. Using ids,
    either explicitly like ``id:foo`` or by using the `default locator
    strategy` simply like ``foo``, is recommended when possible, because
    the syntax is simple and locating elements by an id is fast for browsers.
    If an element does not have an id or the id is not stable, other
    solutions need to be used. If an element has a unique tag name or class,
    using ``tag``, ``class`` or ``css`` strategy like ``tag:h1``,
    ``class:example`` or ``css:h1.example`` is often an easy solution. In
    more complex cases using XPath expressions is typically the best
    approach. They are very powerful but a downside is that they can also
    get complex.

    Examples:

    | `Click Element` | id:foo                      | # Element with id 'foo'. |
    | `Click Element` | css:div#foo h1              | # h1 element under div with id 'foo'. |
    | `Click Element` | xpath: //div[@id="foo"]//h1 | # Same as the above using XPath, not CSS. |
    | `Click Element` | xpath: //*[contains(text(), "example")] | # Element containing text 'example'. |

    *NOTE:*

    - Using the ``sizzle`` strategy or its alias ``jquery`` requires that
      the system under test contains the jQuery library.
    

    === Implicit XPath strategy ===

    If the locator starts with ``//`` or ``(//``, the locator is considered
    to be an XPath expression. In other words, using ``//div`` is equivalent
    to using explicit ``xpath://div``.

    Examples:

    | `Click Element` | //div[@id="foo"]//h1 |
    | `Click Element` | (//div)[2]           |

    
    == Using WebElements ==

    In addition to specifying a locator as a string, it is possible to use
    Selenium's WebElement objects. This requires first getting a WebElement,
    for example, by using the `Get WebElement` keyword.

    | ${elem} =       | `Get WebElement` | id:example |
    | `Click Element` | ${elem}          |            |

    == Custom locators ==

    If more complex lookups are required than what is provided through the
    default locators, custom lookup strategies can be created. Using custom
    locators is a two part process. First, create a keyword that returns
    a WebElement that should be acted on:

    | Custom Locator Strategy | [Arguments] | ${browser} | ${locator} | ${tag} | ${constraints} |
    |   | ${element}= | Execute Javascript | return window.document.getElementById('${locator}'); |
    |   | [Return] | ${element} |

    This keyword is a reimplementation of the basic functionality of the
    ``id`` locator where ``${browser}`` is a reference to a WebDriver
    instance and ``${locator}`` is name of the locator strategy. To use
    this locator it must first be registered by using the
    `Add Location Strategy` keyword:

    | `Add Location Strategy` | custom | Custom Locator Strategy |

    The first argument of `Add Location Strategy` specifies the name of
    the strategy and it must be unique. After registering the strategy,
    the usage is the same as with other locators:

    | `Click Element` | custom:example |

    See the `Add Location Strategy` keyword for more details.

    = Timeouts, waits and delays =

    This section discusses different ways how to wait for elements to
    appear on web pages and to slow down execution speed otherwise.
    It also explains the `time format` that can be used when setting various
    timeouts, waits and delays.

    == Timeout ==

    rbfWEB contains various keywords that have an optional
    ``timeout`` argument that specifies how long these keywords should
    wait for certain events or actions. These keywords include, for example,
    ``Wait ...`` keywords and keywords related to alerts. Additionally
    `Execute Async Javascript`. although it does not have ``timeout``,
    argument, uses timeout to define how long asynchronous JavaScript
    can run.

    The default timeout these keywords use can be set globally either by
    using the `Set Selenium Timeout` keyword or with the ``timeout`` argument
    when `importing` the library. See `time format` below for supported
    timeout syntax.

    == Implicit wait ==

    Implicit wait specifies the maximum time how long Selenium waits when
    searching for elements. It can be set by using the `Set Selenium Implicit
    Wait` keyword or with the ``implicit_wait`` argument when `importing`
    the library. See [http://seleniumhq.org/docs/04_webdriver_advanced.html|
    Selenium documentation] for more information about this functionality.

    See `time format` below for supported syntax.

    == Selenium speed ==

    Selenium execution speed can be slowed down globally by using `Set
    Selenium speed` keyword. This functionality is designed to be used for
    demonstrating or debugging purposes. Using it to make sure that elements
    appear on a page is not a good idea, and the above explained timeouts
    and waits should be used instead.

    See `time format` below for supported syntax.

    == Time format ==

    All timeouts and waits can be given as numbers considered seconds
    (e.g. ``0.5`` or ``42``) or in Robot Framework's time syntax
    (e.g. ``1.5 seconds`` or ``1 min 30 s``). For more information about
    the time syntax see the
    [http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#time-format|Robot Framework User Guide].

    = Run-on-failure functionality =

    rbfWEB has a handy feature that it can automatically execute
    a keyword if any of its own keywords fails. By default it uses the
    `Capture Page Screenshot` keyword, but this can be changed either by
    using the `Register Keyword To Run On Failure` keyword or with the
    ``run_on_failure`` argument when `importing` the library. It is
    possible to use any keyword from any imported library or resource file.

    The run-on-failure functionality can be disabled by using a special
    value ``NOTHING`` or anything considered false (see `Boolean arguments`)
    such as ``NONE``.

    = Boolean arguments =

    Some keywords accept arguments that are handled as Boolean values true or
    false. If such an argument is given as a string, it is considered false if
    it is either empty or case-insensitively equal to ``false``, ``no``, ``off``,
     ``0`` or ``none``. Other strings are considered true regardless their value, and
    other argument types are tested using same
    [https://docs.python.org/2/library/stdtypes.html#truth-value-testing|rules as in Python].

    True examples:

    | `Set Screenshot Directory` | ${RESULTS} | persist=True    | # Strings are generally true.    |
    | `Set Screenshot Directory` | ${RESULTS} | persist=yes     | # Same as the above.             |
    | `Set Screenshot Directory` | ${RESULTS} | persist=${TRUE} | # Python True is true.           |
    | `Set Screenshot Directory` | ${RESULTS} | persist=${42}   | # Numbers other than 0 are true. |

    False examples:

    | `Set Screenshot Directory` | ${RESULTS} | persist=False    | # String false is false.        |
    | `Set Screenshot Directory` | ${RESULTS} | persist=no       | # Also string no is false.      |
    | `Set Screenshot Directory` | ${RESULTS} | persist=NONE     | # String NONE is false.         |
    | `Set Screenshot Directory` | ${RESULTS} | persist=${EMPTY} | # Empty string is false.        |
    | `Set Screenshot Directory` | ${RESULTS} | persist=${FALSE} | # Python False is false.        |
    | `Set Screenshot Directory` | ${RESULTS} | persist=${NONE}  | # Python None is false.         |

    
    = Thread support =

    rbfWEB is not thread safe. This is mainly due because the underlying
    [https://github.com/SeleniumHQ/selenium/wiki/Frequently-Asked-Questions#q-is-webdriver-thread-safe|
    Selenium tool is not thread safe] within one browser/driver instance.
    Because of the limitation in the Selenium side, the keywords or the
    API provided the rbfWEB is not thread safe.
    """
    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__

    def __init__(self, timeout=5.0, implicit_wait=0.0,
                 run_on_failure='Capture Page Screenshot',
                 screenshot_root_directory=None):
        """rbfWEB can be imported with several optional arguments.

        - ``timeout``:
          Default value for `timeouts` used with ``Wait ...`` keywords.
        - ``implicit_wait``:
          Default value for `implicit wait` used when locating elements.
        - ``run_on_failure``:
          Default action for the `run-on-failure functionality`.
        - ``screenshot_root_directory``:
          Location where possible screenshots are created. If not given,
          the directory where the log file is written is used.
        """
        self.timeout = timestr_to_secs(timeout)
        self.implicit_wait = timestr_to_secs(implicit_wait)
        self.speed = 0.0
        self.run_on_failure_keyword \
            = RunOnFailureKeywords.resolve_keyword(run_on_failure)
        self._running_on_failure_keyword = False
        self.screenshot_root_directory = screenshot_root_directory
        libraries = [
            AlertKeywords(self),
            BrowserManagementKeywords(self),
            CookieKeywords(self),
            ElementKeywords(self),
            FormElementKeywords(self),
            FrameKeywords(self),
            JavaScriptKeywords(self),
            RunOnFailureKeywords(self),
            ScreenshotKeywords(self),
            SelectElementKeywords(self),
            TableElementKeywords(self),
            WaitingKeywords(self),
            WindowKeywords(self),
            NettbankKeywords(self)
        ]
        self._drivers = WebDriverCache()
        DynamicCore.__init__(self, libraries)
        self.ROBOT_LIBRARY_LISTENER = LibraryListener()
        self._element_finder = ElementFinder(self)

    _speed_in_secs = Deprecated('_speed_in_secs', 'speed')
    _timeout_in_secs = Deprecated('_timeout_in_secs', 'timeout')
    _implicit_wait_in_secs = Deprecated('_implicit_wait_in_secs',
                                        'implicit_wait')
    _run_on_failure_keyword = Deprecated('_run_on_failure_keyword',
                                         'run_on_failure_keyword')

    def run_keyword(self, name, args, kwargs):
        try:
            return DynamicCore.run_keyword(self, name, args, kwargs)
        except Exception:
            self.failure_occurred()
            raise

    def register_driver(self, driver, alias):
        """Add's a `driver` to the library WebDriverCache.

        :param driver: Instance of the Selenium `WebDriver`.
        :type driver: selenium.webdriver.remote.webdriver.WebDriver
        :param alias: Alias given for this `WebDriver` instance.
        :type alias: str
        :return: The index of the `WebDriver` instance.
        :rtype: int
        """
        return self._drivers.register(driver, alias)

    def failure_occurred(self):
        """Method that is executed when a rbfWEB keyword fails.

        By default executes the registered run-on-failure keyword.
        Libraries extending rbfWEB can overwrite this hook
        method if they want to provide custom functionality instead.
        """
        if self._running_on_failure_keyword or not self.run_on_failure_keyword:
            return
        try:
            self._running_on_failure_keyword = True
            BuiltIn().run_keyword(self.run_on_failure_keyword)
        except Exception as err:
            logger.warn("Keyword '%s' could not be run on failure: %s"
                        % (self.run_on_failure_keyword, err))
        finally:
            self._running_on_failure_keyword = False

    @property
    def driver(self):
        """Current active driver.

        :rtype: selenium.webdriver.remote.webdriver.WebDriver
        :raises rbfWEB.errors.NoOpenBrowser: If browser is not open.
        """
        if not self._drivers.current:
            raise NoOpenBrowser('No browser is open.')
        return self._drivers.current

    def find_element(self, locator, parent=None):
        """Find element matching `locator`.

        :param locator: Locator to use when searching the element.
            See library documentation for the supported locator syntax.
        :type locator: str or selenium.webdriver.remote.webelement.WebElement
        :param parent: Optional parent `WebElememt` to search child elements
            from. By default search starts from the root using `WebDriver`.
        :type parent: selenium.webdriver.remote.webelement.WebElement
        :return: Found `WebElement`.
        :rtype: selenium.webdriver.remote.webelement.WebElement
        :raises rbfWEB.errors.ElementNotFound: If element not found.
        """
        return self._element_finder.find(locator, parent=parent)

    def find_elements(self, locator, parent=None):
        """Find all elements matching `locator`.

        :param locator: Locator to use when searching the element.
            See library documentation for the supported locator syntax.
        :type locator: str or selenium.webdriver.remote.webelement.WebElement
        :param parent: Optional parent `WebElememt` to search child elements
            from. By default search starts from the root using `WebDriver`.
        :type parent: selenium.webdriver.remote.webelement.WebElement
        :return: list of found `WebElement` or e,mpty if elements are not found.
        :rtype: list[selenium.webdriver.remote.webelement.WebElement]
        """
        return self._element_finder.find(locator, first_only=False,
                                         required=False, parent=parent)

    @property
    def _cache(self):
        warnings.warn('"rbfWEB._cache" is deprecated, '
                      'use public API instead.', DeprecationWarning)
        return self._drivers

    def _current_browser(self):
        warnings.warn('"rbfWEB._current_browser" is deprecated, '
                      'use "rbfWEB.driver" instead.',
                      DeprecationWarning)
        return self.driver

    def _run_on_failure(self):
        warnings.warn('"rbfWEB._run_on_failure" is deprecated, '
                      'use "rbfWEB.failure_occurred" instead.',
                      DeprecationWarning)
        self.failure_occurred()



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import namedtuple

from robot.utils import plural_or_not, seq2str

from rbfWEB.base import LibraryComponent, keyword


class JavaScriptKeywords(LibraryComponent):

    js_marker = 'JAVASCRIPT'
    arg_marker = 'ARGUMENTS'

    @keyword
    def execute_javascript(self, *code):
        """Executes the given JavaScript code with possible arguments.

        ``code`` may be divided into multiple cells in the test data and
        ``code`` may contain multiple lines of code and arguments. In that case,
        the JavaScript code parts are concatenated together without adding
        spaces and optional arguments are separated from ``code``.

        If ``code`` is a path to an existing file, the JavaScript
        to execute will be read from that file. Forward slashes work as
        a path separator on all operating systems.

        The JavaScript executes in the context of the currently selected
        frame or window as the body of an anonymous function. Use ``window``
        to refer to the window of your application and ``document`` to refer
        to the document object of the current frame or window, e.g.
        ``document.getElementById('example')``.

        This keyword returns whatever the executed JavaScript code returns.
        Return values are converted to the appropriate Python types.

        It is possible to provide JavaScript
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver_remote/selenium.webdriver.remote.webdriver.html#selenium.webdriver.remote.webdriver.WebDriver.execute_script|
        arguments] as part of ``code`` argument. The JavaScript code and
        arguments must be separated with `JAVASCRIPT` and `ARGUMENTS` markers
        and must used exactly with this format. If the Javascript code is
        first, then the `JAVASCRIPT` marker is optional. The order of
        `JAVASCRIPT` and `ARGUMENTS` markers can swapped, but if `ARGUMENTS`
        is first marker, then `JAVASCRIPT` marker is mandatory. It is only
        allowed to use `JAVASCRIPT` and `ARGUMENTS` markers only one time in the
        ``code`` argument.

        Examples:
        | `Execute JavaScript` | window.myFunc('arg1', 'arg2') |
        | `Execute JavaScript` | ${CURDIR}/js_to_execute.js    |
        | `Execute JavaScript` | alert(arguments[0]); | ARGUMENTS | 123 |
        | `Execute JavaScript` | ARGUMENTS | 123 | JAVASCRIPT | alert(arguments[0]); |
        """
        js_code, js_args = self._get_javascript_to_execute(code)
        self._js_logger('Executing JavaScript', js_code, js_args)
        return self.driver.execute_script(js_code, *js_args)

    @keyword
    def execute_async_javascript(self, *code):
        """Executes asynchronous JavaScript code with possible arguments.

        Similar to `Execute Javascript` except that scripts executed with
        this keyword must explicitly signal they are finished by invoking the
        provided callback. This callback is always injected into the executed
        function as the last argument.

        Scripts must complete within the script timeout or this keyword will
        fail. See the `Timeout` section for more information.

        It is possible to provide JavaScript
        [https://seleniumhq.github.io/selenium/docs/api/py/webdriver_remote/selenium.webdriver.remote.webdriver.html#selenium.webdriver.remote.webdriver.WebDriver.execute_async_script|
        arguments] as part of ``code`` argument. See `Execute Javascript` for
        more details.

        Examples:
        | `Execute Async JavaScript` | var callback = arguments[arguments.length - 1]; window.setTimeout(callback, 2000); |
        | `Execute Async JavaScript` | ${CURDIR}/async_js_to_execute.js |
        | ${result} = | `Execute Async JavaScript`                      |
        | ...         | var callback = arguments[arguments.length - 1]; |
        | ...         | function answer(){callback("text");};           |
        | ...         | window.setTimeout(answer, 2000);                |
        | `Should Be Equal` | ${result} | text |
        """
        js_code, js_args = self._get_javascript_to_execute(code)
        self._js_logger('Executing Asynchronous JavaScript', js_code, js_args)
        return self.driver.execute_async_script(js_code, *js_args)

    def _js_logger(self, base, code, args):
        message = '%s:\n%s\n' % (base, code)
        if args:
            message = ('%sBy using argument%s:\n%s'
                       % (message, plural_or_not(args), seq2str(args)))
        else:
            message = '%sWithout any arguments.' % message
        self.info(message)

    def _get_javascript_to_execute(self, code):
        js_code, js_args = self._separate_code_and_args(code)
        if not js_code:
            raise ValueError('JavaScript code was not found from code argument.')
        js_code = ''.join(js_code)
        path = js_code.replace('/', os.sep)
        if os.path.isfile(path):
            js_code = self._read_javascript_from_file(path)
        return js_code, js_args

    def _separate_code_and_args(self, code):
        code = list(code)
        self._check_marker_error(code)
        index = self._get_marker_index(code)
        if self.arg_marker not in code:
            return code[index.js + 1:], []
        if self.js_marker not in code:
            return code[0:index.arg], code[index.arg + 1:]
        else:
            if index.js == 0:
                return code[index.js + 1:index.arg], code[index.arg + 1:]
            else:
                return code[index.js + 1:], code[index.arg + 1:index.js]

    def _check_marker_error(self, code):
        if not code:
            raise ValueError('There must be at least one argument defined.')
        message = None
        template = '%s marker was found two times in the code.'
        if code.count(self.js_marker) > 1:
            message = template % self.js_marker
        if code.count(self.arg_marker) > 1:
            message = template % self.arg_marker
        index = self._get_marker_index(code)
        if index.js > 0 and index.arg != 0:
            message = template % self.js_marker
        if message:
            raise ValueError(message)

    def _get_marker_index(self, code):
        Index = namedtuple('Index', 'js arg')
        if self.js_marker in code:
            js = code.index(self.js_marker)
        else:
            js = -1
        if self.arg_marker in code:
            arg = code.index(self.arg_marker)
        else:
            arg = -1
        return Index(js=js, arg=arg)

    def _read_javascript_from_file(self, path):
        self.info('Reading JavaScript from file <a href="file://%s">%s</a>.'
                  % (path.replace(os.sep, '/'), path), html=True)
        with open(path) as file:
            return file.read().strip()



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import csv


class CsvReader:

    def __init__(self,
                 file,
                 encoding,
                 testcase_table_name,
                 dialect='Excel-EU',
                 delimiter=';',
                 quotechar='"',
                 escapechar='\\',
                 doublequote=True,
                 skipinitialspace=False,
                 lineterminator='\r\n'
                 ):

        self.file = file
        self.csv_encoding = encoding
        self.csv_dialect = dialect
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace
        self.lineterminator = lineterminator

        self.TESTCASE_TABLE_NAME = testcase_table_name
        self.data_table = None
        self.index = None

    def get_data_from_csv(self):
        self._register_dialects()
        self._read_file_to_dictionaries()
        return self.data_table

    def _register_dialects(self):
        if self.csv_dialect == 'UserDefined':
            csv.register_dialect(self.csv_dialect,
                                 delimiter=self.delimiter,
                                 quotechar=self.quotechar,
                                 escapechar=self.escapechar,
                                 doublequote=self.doublequote,
                                 skipinitialspace=self.skipinitialspace,
                                 lineterminator=self.lineterminator,
                                 quoting=csv.QUOTE_ALL)
        elif self.csv_dialect == 'Excel-EU':
            csv.register_dialect(self.csv_dialect,
                                 delimiter=';',
                                 quotechar='"',
                                 escapechar='\\',
                                 doublequote=True,
                                 skipinitialspace=False,
                                 lineterminator='\r\n',
                                 quoting=csv.QUOTE_ALL)

    def _read_file_to_dictionaries(self):
        with open(self.file, 'r', encoding=self.csv_encoding) as csvfile:
            reader = csv.reader(csvfile, self.csv_dialect)
            table = {}
            header = []
            for row_index, row in enumerate(reader):
                if row_index == 0:
                    header = row
                    if header[0] == self.TESTCASE_TABLE_NAME:
                        for cell in header:
                            table[cell] = []
                    else:
                        raise SyntaxError('First column is not "'
                                          + self.TESTCASE_TABLE_NAME
                                          + '". This Column is mandatory.')
                else:
                    for cell_index, cell in enumerate(row):
                        table[header[cell_index]].append(cell)
            self.data_table = table



from TestStack.White.UIItems import Button, CheckBox, RadioButton
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword
from rbfDesktop.utils.click import Clicks


class ButtonKeywords(LibraryComponent):
    @keyword
    def click_button(self, locator, x_offset=0, y_offset=0):
        """Clicks a button.

        ``locator`` is the locator of the button or Button item object.
        Locator syntax is explained in `Item locators`.

        Optional arguments ``x_offset`` and ``y_offset`` can be used to define the coordinates to click at,
        relative to the center of the item.
        """
        button = self.state._get_typed_item_by_locator(Button, locator)
        Clicks.click(button, x_offset, y_offset)

    @keyword
    def button_text_should_be(self, locator, expected_text, case_sensitive=True):
        """Verifies exact text in a button.

        ``locator`` is the locator of the button or Button item object.
        Locator syntax is explained in `Item locators`.

        ``expected_text`` is the expected button text.

        ``case_sensitive`` defines if the text comparison is case sensitive (True/False). Defaults to ``True``.
        Boolean values are evaluated in the same way as the Robot Framework BuiltIn library does, see
        [http://robotframework.org/robotframework | the documentation of BuiltIn] for details.
        """
        button = self.state._get_typed_item_by_locator(Button, locator)
        self.state._verify_string_value(expected_text, button.Text, case_sensitive)

    @keyword
    def button_text_should_contain(self, locator, expected_text, case_sensitive=True):
        """Verifies expected text is found in a button.

        ``locator`` is the locator of the button or Button item object.
        Locator syntax is explained in `Item locators`.

        ``expected_text`` is the expected button text.

        ``case_sensitive`` defines if the text comparison is case sensitive (True/False). Defaults to ``True``.
        Boolean values are evaluated in the same way as the Robot Framework BuiltIn library does, see
        [http://robotframework.org/robotframework | the documentation of BuiltIn] for details.
        """
        button = self.state._get_typed_item_by_locator(Button, locator)
        self.state._contains_string_value(expected_text, button.Text, case_sensitive)

    @keyword
    def verify_button(self, locator, expected):
        """*DEPRECATED.* Use `Button Text Should Be` or `Button Text Should Contain` instead.

        Verifies text in a button.

        ``locator`` is the locator of the button or Button item object.

        ``expected`` is the expected button text.
        """
        button = self.state._get_typed_item_by_locator(Button, locator)
        self.state._verify_value(expected, button.Text)

    @keyword
    def select_radio_button(self, locator):
        """Selects a radio button.

        ``locator`` is the locator of the radio button or RadioButton item object.
        Locator syntax is explained in `Item locators`.
        """
        radiobutton = self.state._get_typed_item_by_locator(RadioButton, locator)
        radiobutton.Select()

    @keyword
    def verify_radio_button(self, locator, expected):
        """Verifies state of a radio button.

        ``locator`` is the locator of the radio button or RadioButton item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected state (True/False).
        """
        radiobutton = self.state._get_typed_item_by_locator(RadioButton, locator)
        self.state._verify_value(bool(expected), radiobutton.IsSelected)

    @keyword
    def get_radio_button_state(self, locator):
        """Returns the state of a radio button.

        Returns `True` if the radio button is selected, `False` if not.

        ``locator`` is the locator of the radio button or RadioButton item object.
        Locator syntax is explained in `Item locators`.
        """
        radiobutton = self.state._get_typed_item_by_locator(RadioButton, locator)
        return radiobutton.IsSelected

    @keyword
    def toggle_check_box(self, locator):
        """Toggles a check box.

        ``locator`` is the locator of the check box or CheckBox item object.
        Locator syntax is explained in `Item locators`.
        """
        checkbox = self.state._get_typed_item_by_locator(CheckBox, locator)
        checkbox.Toggle()

    @keyword
    def verify_check_box(self, locator, expected):
        """Verifies state of a check box.

        ``locator`` is the locator of the check box or CheckBox item object.
        Locator syntax is explained in `Item locators`.

        ``expected`` is the expected state (True/False).
        """
        checkbox = self.state._get_typed_item_by_locator(CheckBox, locator)
        self.state._verify_value(bool(expected), checkbox.IsSelected)

    @keyword
    def get_check_box_state(self, locator):
        """Returns the state of a check box.

        Returns `True` if the check box is selected, `False` if not.

        ``locator`` is the locator of the check box or CheckBox item object.
        Locator syntax is explained in `Item locators`.
        """
        checkbox = self.state._get_typed_item_by_locator(CheckBox, locator)
        return checkbox.IsSelected



import os
import robot
from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn, RobotNotRunningError
from robot.utils import get_link_path, is_truthy
from rbfDesktop.keywords.librarycomponent import LibraryComponent
from rbfDesktop.keywords.robotlibcore import keyword, PY2
from System.Drawing.Imaging import ImageFormat
from TestStack.White import Desktop


class ScreenshotKeywords(LibraryComponent):
    def __init__(self, state, directory):
        super(ScreenshotKeywords, self).__init__(state)
        self.state.screenshooter = self
        self._screenshot_directory = directory

    @keyword
    def set_screenshot_directory(self, path):
        """Sets the directory where rbfDesktop stores screenshots and returns the previously set path.

        If the given directory does not exist in the system, it will be created when the first screenshot is taken.

        ``path`` is the directory path.

        The directory can also be set when the library is imported, see `Importing` for details.
        """
        old_dir = self._screenshot_directory
        self._screenshot_directory = None if path is None else robot.utils.abspath(path)
        return old_dir

    @keyword
    def take_desktop_screenshot(self):
        """Takes a screenshot of the whole desktop and inserts screenshot link to log file.

        Returns path to the screenshot file.
        """
        filepath = self._get_screenshot_path("whitelib_screenshot_{index}.png")
        directory = os.path.dirname(filepath)
        if not os.path.exists(directory):
            os.makedirs(directory)
        logger.info(get_link_path(filepath, self._log_directory))
        logger.info(
            '</td></tr><tr><td colspan="3">'
            '<a href="{src}"><img src="{src}" width="800px"></a>'.format(
                src=get_link_path(filepath, self._log_directory)
            ),
            html=True,
        )
        bmp = Desktop.CaptureScreenshot()
        bmp.Save(filepath, ImageFormat.Png)
        return filepath

    @keyword
    def take_screenshots_on_failure(self, status):
        """Disables or enables automatic screenshot capturing on failure.

        ``status`` is the desired state (True/False) of automatic screenshot on failure.
        Boolean values are evaluated in the same way as the Robot Framework BuiltIn library does, see
        [http://robotframework.org/robotframework | the documentation of BuiltIn] for details.

        rbfDesktop automatically takes a screenshot on failure unless it is disabled with this keyword.
        """
        if is_truthy(status):
            self.state.screenshots_enabled = True
        else:
            self.state.screenshots_enabled = False

    @property
    def _log_directory(self):
        try:
            logfile = BuiltIn().get_variable_value("${LOG FILE}")
            if logfile is None:
                return BuiltIn().get_variable_value("${OUTPUTDIR}")
            return os.path.dirname(logfile)
        except RobotNotRunningError:
            return os.getcwdu() if PY2 else os.getcwd()  # pylint: disable=no-member

    def _get_screenshot_path(self, filename):
        if self._screenshot_directory is not None:
            directory = self._screenshot_directory
        else:
            directory = self._log_directory
        filename = filename.replace("/", os.sep)
        index = 0
        while True:
            index += 1
            formatted = filename.format(index=index)
            filepath = os.path.join(directory, formatted)
            # filename didn't contain {index} or unique path was found
            if formatted == filename or not os.path.exists(filepath):
                return filepath



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rbfWEB.utils import escape_xpath_value


class ContextAware(object):

    def __init__(self, ctx):
        """Base class exposing attributes from the common context.

        :param ctx: The library itself as a context object.
        :type ctx: rbfWEB.rbfWEB
        """
        self.ctx = ctx

    @property
    def driver(self):
        return self.ctx.driver

    @property
    def drivers(self):
        return self.ctx._drivers

    @property
    def element_finder(self):
        return self.ctx._element_finder

    def find_element(self, locator, tag=None, required=True, parent=None):
        """Find element matching `locator`.

        :param locator: Locator to use when searching the element.
            See library documentation for the supported locator syntax.
        :type locator: str or selenium.webdriver.remote.webelement.WebElement
        :param tag: Limit searching only to these elements.
        :type tag: str
        :param required: Raise `ElementNotFound` if element not found when
            true, return `None` otherwise.
        :type required: True or False
        :param parent: Optional parent `WebElememt` to search child elements
            from. By default search starts from the root using `WebDriver`.
        :type parent: selenium.webdriver.remote.webelement.WebElement
        :return: Found `WebElement` or `None` if element not found and
            `required` is false.
        :rtype: selenium.webdriver.remote.webelement.WebElement
        :raises rbfWEB.errors.ElementNotFound: If element not found
            and `required` is true.
        """
        return self.element_finder.find(locator, tag, True, required, parent)

    def find_elements(self, locator, tag=None, parent=None):
        """Find all elements matching `locator`.

        :param locator: Locator to use when searching the element.
            See library documentation for the supported locator syntax.
        :type locator: str or selenium.webdriver.remote.webelement.WebElement
        :param tag: Limit searching only to these elements.
        :type tag: str
        :param parent: Optional parent `WebElememt` to search child elements
            from. By default search starts from the root using `WebDriver`.
        :type parent: selenium.webdriver.remote.webelement.WebElement
        :return: list of found `WebElement` or empty if elements are not found.
        :rtype: list[selenium.webdriver.remote.webelement.WebElement]
        """
        return self.element_finder.find(locator, tag, False, False, parent)

    def is_text_present(self, text):
        locator = "xpath://*[contains(., %s)]" % escape_xpath_value(text)
        return self.find_element(locator, required=False) is not None

    def is_element_enabled(self, locator, tag=None):
        element = self.find_element(locator, tag)
        return (element.is_enabled() and
                element.get_attribute('readonly') is None)

    def is_visible(self, locator):
        element = self.find_element(locator, required=False)
        return element.is_displayed() if element else None



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pandas as pd
import numpy as np


class XlsReader:

    def __init__(self, file):
        self.file = file

    def get_data_from_xls(self):
        data_frame = pd.read_excel(self.file, dtype=str).replace(np.nan, '', regex=True)
        table = {}
        for header in data_frame.columns:
            table[header] = data_frame[header].tolist()
        return table



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function

import os
import sys
import glob
import argparse
import imp

from .common import SuiteRule, ResourceRule, TestRule, KeywordRule, GeneralRule, Rule
from .common import ERROR, WARNING, IGNORE
from .version import __version__
from .parser import RobotFactory, SuiteFile, ResourceFile
from .exceptions import UnknownRuleException

from robot.utils.argumentparser import ArgFileParser


class RfLint(object):
    """Robot Framework Linter"""

    def __init__(self):
        here = os.path.abspath(os.path.dirname(__file__))
        builtin_rules = os.path.join(here, "rules")
        site_rules = os.path.join(here, "site-rules")

        # mapping of class names to instances, to enable us to
        # instantiate each rule exactly once
        self._rules = {}

        for path in (builtin_rules, site_rules):
            for filename in glob.glob(path+"/*.py"):
                if filename.endswith(".__init__.py"):
                    continue
                self._load_rule_file(filename)

    @property
    def suite_rules(self):
        return self._get_rules(SuiteRule)

    @property 
    def resource_rules(self):
        return self._get_rules(ResourceRule)

    @property
    def testcase_rules(self):
        return self._get_rules(TestRule)

    @property
    def keyword_rules(self):
        return self._get_rules(KeywordRule)

    @property
    def general_rules(self):
        return self._get_rules(GeneralRule)

    @property
    def all_rules(self):
        all = self.suite_rules + self.resource_rules + self.testcase_rules + self.keyword_rules + self.general_rules
        return all

    def run(self, args):
        """Parse command line arguments, and run rflint"""

        self.args = self.parse_and_process_args(args)

        if self.args.version:
            print(__version__)
            return 0
            
        if self.args.rulefile:
            for filename in self.args.rulefile:
                self._load_rule_file(filename)

        if self.args.list:
            self.list_rules()
            return 0
        
        if self.args.describe:
            self._describe_rules(self.args.args)
            return 0

        self.counts = { ERROR: 0, WARNING: 0, "other": 0}
            
        for filename in self.args.args:
            if not (os.path.exists(filename)):
                sys.stderr.write("rflint: %s: No such file or directory\n" % filename)
                continue
            if os.path.isdir(filename):
                self._process_folder(filename)
            else:
                self._process_file(filename)

        if self.counts[ERROR] > 0:
            return self.counts[ERROR] if self.counts[ERROR] < 254 else 255

        return 0

    def _is_valid_rule(self, rule_name):
        for rule in self.all_rules:
            if rule_name.lower() == rule.name.lower():
                return True
        return False

    def _describe_rules(self, rule_names):
        for rulename in rule_names:
            if not self._is_valid_rule(rulename):
                raise UnknownRuleException(rulename)

        requested_rules = [rule.strip().lower() for rule in rule_names]
        for rule in sorted(self.all_rules, key=lambda rule: rule.name):
            if rule.name.lower() in requested_rules or len(requested_rules) == 0:
                print(rule.name)
                for line in rule.doc.split("\n"):
                    print("    " + line)

    def _process_folder(self, path):
        if self.args.recursive:
            for root, subdirs, filenames in os.walk(path):
                self._process_files(root, filenames)
        else:
            self._process_files(path, os.listdir(path))

    def _process_files(self, folder, filenames):
        for filename in filenames:
            name, ext = os.path.splitext(filename)
            if ext.lower() in (".robot", ".txt", ".tsv"):
                self._process_file(os.path.join(folder, filename))
 
    def _process_file(self, filename):
        # this is used by the reporting mechanism to know if it
        # should print the filename. Once it has been printed it
        # will be reset so that it won't get printed again until 
        # we process the next file.
        self._print_filename = filename if self.args.print_filenames else None

        robot_file = RobotFactory(filename)
        for rule in self.general_rules:
            if rule.severity != IGNORE:
                rule.apply(robot_file)

        if isinstance(robot_file, SuiteFile):
            for rule in self.suite_rules:
                if rule.severity != IGNORE:
                    rule.apply(robot_file)
            for testcase in robot_file.testcases:
                for rule in self.testcase_rules:
                    if rule.severity != IGNORE:
                        rule.apply(testcase)

        if isinstance(robot_file, ResourceFile):
            for rule in self.resource_rules:
                if rule.severity != IGNORE:
                    rule.apply(robot_file)

        for keyword in robot_file.keywords:
            for rule in self.keyword_rules:
                if rule.severity != IGNORE:
                    rule.apply(keyword)

    def list_rules(self):
        """Print a list of all rules"""
        for rule in sorted(self.all_rules, key=lambda rule: rule.name):
            print(rule)
            if self.args.verbose:
                for line in rule.doc.split("\n"):
                    print("    ", line)

    def report(self, linenumber, filename, severity, message, rulename, char):
        """Report a rule violation"""

        if self._print_filename is not None:
            # we print the filename only once. self._print_filename
            # will get reset each time a new file is processed.
            print("+ " + self._print_filename)
            self._print_filename = None

        if severity in (WARNING, ERROR):
            self.counts[severity] += 1
        else:
            self.counts["other"] += 1

        print(self.args.format.format(linenumber=linenumber, filename=filename,
                                      severity=severity, message=message.encode('utf-8'),
                                      rulename=rulename, char=char))

    def _get_rules(self, cls):
        """Returns a list of rules of a given class
        
        Rules are treated as singletons - we only instantiate each
        rule once. 
        """

        result = []
        for rule_class in cls.__subclasses__():
            rule_name = rule_class.__name__.lower()
            if rule_name not in self._rules:
                rule = rule_class(self)
                self._rules[rule_name] = rule
            result.append(self._rules[rule_name])
        return result

    def _load_rule_file(self, filename):
        """Import the given rule file"""
        if not (os.path.exists(filename)):
            sys.stderr.write("rflint: %s: No such file or directory\n" % filename)
            return
        try:
            basename = os.path.basename(filename)
            (name, ext) = os.path.splitext(basename)
            imp.load_source(name, filename)
        except Exception as e:
            sys.stderr.write("rflint: %s: exception while loading: %s\n" % (filename, str(e)))

    def parse_and_process_args(self, args):
        """Handle the parsing of command line arguments."""

        parser = argparse.ArgumentParser(
            prog="python -m rflint",
            description="A style checker for robot framework plain text files.",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog = (
                "You can use 'all' in place of RULENAME to refer to all rules. \n"
                "\n"
                "For example: '--ignore all --warn DuplicateTestNames' will ignore all\n"
                "rules except DuplicateTestNames.\n"
                "\n"
                "FORMAT is a string that performs a substitution on the following \n"
                "patterns: {severity}, {linenumber}, {char}, {message}, and {rulename}.\n"
                "\n"
                "For example: --format 'line: {linenumber}: message: {message}'. \n"
                "\n"
                "ARGUMENTFILE is a filename with contents that match the format of \n"
                "standard robot framework argument files\n"
                "\n"
                "If you give a directory as an argument, all files in the directory\n"
                "with the suffix .txt, .robot or .tsv will be processed. With the \n"
                "--recursive option, subfolders within the directory will also be\n"
                "processed."
                )
            )
        parser.add_argument("--error", "-e", metavar="RULENAME", action=SetErrorAction,
                            help="Assign a severity of ERROR to the given RULENAME")
        parser.add_argument("--ignore", "-i", metavar="RULENAME", action=SetIgnoreAction,
                            help="Ignore the given RULENAME")
        parser.add_argument("--warning", "-w", metavar="RULENAME", action=SetWarningAction,
                            help="Assign a severity of WARNING for the given RULENAME")
        parser.add_argument("--list", "-l", action="store_true",
                            help="show a list of known rules and exit")
        parser.add_argument("--describe", "-d", action="store_true",
                            help="describe the given rules")
        parser.add_argument("--no-filenames", action="store_false", dest="print_filenames", 
                            default=True,
                            help="suppress the printing of filenames")
        parser.add_argument("--format", "-f", 
                            help="Define the output format",
                            default='{severity}: {linenumber}, {char}: {message} ({rulename})')
        parser.add_argument("--version", action="store_true", default=False,
                            help="Display version number and exit")
        parser.add_argument("--verbose", "-v", action="store_true", default=False,
                            help="Give verbose output")
        parser.add_argument("--configure", "-c", action=ConfigureAction,
                            help="Configure a rule")
        parser.add_argument("--recursive", "-r", action="store_true", default=False,
                            help="Recursively scan subfolders in a directory")
        parser.add_argument("--rulefile", "-R", action=RulefileAction,
                            help="import additional rules from the given RULEFILE")
        parser.add_argument("--argumentfile", "-A", action=ArgfileLoader,
                            help="read arguments from the given file")
        parser.add_argument('args', metavar="file", nargs=argparse.REMAINDER)

        # create a custom namespace, in which we can store a reference to
        # our rules. This lets the custom argument actions access the list
        # of rules
        ns = argparse.Namespace()
        setattr(ns, "app", self)
        args = parser.parse_args(args, ns)

        Rule.output_format = args.format

        return args


class RulefileAction(argparse.Action):
    def __call__(self, parser, namespace, arg, option_string=None):
        app = getattr(namespace, "app")
        app._load_rule_file(arg)


class ConfigureAction(argparse.Action):
    def __call__(self, parser, namespace, arg, option_string=None):
        rulename, argstring = arg.split(":", 1)
        args = argstring.split(":")
        app = getattr(namespace, "app")

        for rule in app.all_rules:
            if rulename == rule.name:
                rule.configure(*args)
                return
        raise UnknownRuleException(rulename)


class SetStatusAction(argparse.Action):
    """Abstract class which provides a method for checking the rule name"""
    def check_rule_name(self, rulename, rules):
        if (rulename != "all" and
            rulename.lower() not in [rule.name.lower() for rule in rules]):
            raise UnknownRuleException(rulename)


class SetWarningAction(SetStatusAction):
    """Called when the argument parser encounters --warning"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = WARNING


class SetErrorAction(SetStatusAction):
    """Called when the argument parser encounters --error"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = ERROR


class SetIgnoreAction(SetStatusAction):
    """Called when the argument parser encounters --ignore"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = IGNORE


class ArgfileLoader(argparse.Action):
    """Called when the argument parser encounters --argumentfile"""
    def __call__ (self, parser, namespace, values, option_string = None):
        ap = ArgFileParser(["--argumentfile","-A"])
        args = ap.process(["-A", values])
        parser.parse_args(args, namespace)



import logging

import openpyxl
from rbfExcel.reader import ExcelReader
from rbfExcel.utils import (BoolFormat, DataType, DateFormat, NumberFormat,
                              copy_file, del_file, excel_name2coord,
                              get_file_path, is_file, random_temp_file)
from six import PY2
from xlutils.copy import copy
from xlwt import Workbook, easyxf

LOGGER = logging.getLogger(__name__)


class XlsWriter:
    def __init__(self, workbook=None):
        self.wwb = copy(workbook) if workbook else Workbook(encoding='utf8')
        # Init sheet with new file
        if not workbook:
            self.create_sheet('Sheet')

    def create_sheet(self, sheet_name):
        self.wwb.add_sheet(sheet_name)

    def write_to_cell(self, sheet_name, column, row, value, data_format=None):
        LOGGER.info('XLS Writer...')
        sheet = self.wwb.get_sheet(sheet_name)
        LOGGER.debug('Sheet %s', sheet_name)
        LOGGER.debug('column %s', column)
        LOGGER.debug('row %s', row)
        LOGGER.debug('Type value %s', type(value))
        LOGGER.debug('Data format %s', str(data_format))
        sheet.write(int(row), int(column), label=value, style=easyxf('', num_format_str=data_format))

    def save(self, new_path):
        self.wwb.save(new_path)


class XlsxWriter:
    def __init__(self, file_path, is_new=False):
        LOGGER.debug('Opening writeable file at %s', file_path)
        self.wwb = openpyxl.Workbook() if is_new else openpyxl.load_workbook(file_path, data_only=True)

    def create_sheet(self, sheet_name):
        self.wwb.create_sheet(title=sheet_name)

    def write_to_cell(self, sheet_name, column, row, value, data_format=None):
        LOGGER.info('XLSX Writer...')
        sheet = self.wwb[sheet_name]
        LOGGER.debug('Sheet %s', sheet_name)
        LOGGER.debug('column %s', column)
        LOGGER.debug('row %s', row)
        LOGGER.debug('Type value %s', type(value))
        LOGGER.debug('Data format %s', str(data_format))
        cell = sheet.cell(row=int(row) + 1, column=int(column) + 1, value=value)
        if data_format:
            cell.number_format = data_format

    def save(self, new_path):
        self.wwb.save(new_path)


class ExcelWriter(ExcelReader):

    def __init__(self, file_path, new_path=None, override=False,
                 date_format=DateFormat(), number_format=NumberFormat(), bool_format=BoolFormat()):
        self.is_update = new_path is None or new_path == file_path
        self.new_path = None if self.is_update else get_file_path(new_path)
        self.override = override
        if self.new_path and is_file(self.new_path):
            if self.override:
                del_file(self.new_path)
            else:
                message = 'File ' + self.new_path + ' already existed. Use `override=True` to force override file'
                raise IOError(message) if PY2 else FileExistsError(message)
        try:
            super(ExcelWriter, self).__init__(file_path, date_format, number_format, bool_format)
            self.writer = XlsWriter(self.workbook) if self.is_xls else XlsxWriter(self.file_path)
            self.is_new = False
        except (IOError if PY2 else FileNotFoundError) as _:
            LOGGER.debug('Create new Excel file in %s', self.file_path)
            self.writer = XlsWriter() if self.is_xls else XlsxWriter(self.file_path, is_new=True)
            self.is_new = True

    def save_excel(self):
        """
        Saves the Excel file
        """
        save_path = self.file_path if self.is_new else random_temp_file(ext=self.extension) if self.is_update else self.new_path
        LOGGER.debug('Save Excel file to %s', save_path)
        self.writer.save(save_path)
        if not self.is_new and self.is_update:
            self.workbook.release_resources()
            copy_file(save_path, self.file_path, True)

    def create_sheet(self, sheet_name):
        """
        Creates and appends new Excel worksheet using the new sheet name to the current workbook.
        """
        self.writer.create_sheet(sheet_name)

    def write_to_cell_by_name(self, sheet_name, cell_name, value, data_type=None):
        col, row = excel_name2coord(cell_name)
        self.write_to_cell(sheet_name, col, row, value, data_type)

    def write_to_cell(self, sheet_name, column, row, value, data_type=None):
        """
        Using the sheet name the value of the indicated cell is set to be the value given in the parameter.
        """
        dtype = DataType.parse_type(data_type)
        LOGGER.debug('Write To Sheet: %s - Col: %s - Row: %s', sheet_name, column, row)
        LOGGER.debug('Data Type: %s', dtype)
        LOGGER.debug('Value Type: %s', type(value))
        if DataType.is_bool(dtype, value):
            self.writer.write_to_cell(sheet_name, column, row, self.bool_format.parse(value))
        elif DataType.is_date(dtype, value):
            raw_value = self.date_format.parse(dtype, value)
            dformat = self.date_format.get_excel_format(dtype, value)
            LOGGER.debug('Write Date Value:')
            LOGGER.debug('Raw Type: %s', type(raw_value))
            LOGGER.debug(raw_value)
            LOGGER.debug('Excel Format: %s', dformat)
            LOGGER.debug('-' * 10)
            self.writer.write_to_cell(sheet_name, column, row, raw_value, dformat)
        elif DataType.is_number(dtype, value):
            raw_value = self.number_format.parse(dtype, value)
            nformat = self.number_format.get_excel_format(dtype)
            LOGGER.debug('Raw Type: %s', type(raw_value))
            LOGGER.debug(raw_value)
            LOGGER.debug('Excel Format: %s', nformat)
            LOGGER.debug('-' * 10)
            self.writer.write_to_cell(sheet_name, column, row, raw_value, nformat)
        else:
            self.writer.write_to_cell(sheet_name, column, row, str(value))

    def write_to_cell_by_row_and_column_header(self, sheet_name, row_header, column_header, value, data_type=None):
        """
        Uses the cell value to write the data from that cell.
        """
        print("in writer class")
        sheet = self._get_sheet(sheet_name)
        row = sheet.row(0)
        for colidx, cell in enumerate(row):
            if cell.value == column_header :
                self.colidx = colidx
                        
        col = sheet.col(0)
        for rowidx, cell in enumerate(col):
            if cell.value == row_header :
                self.rowidx = rowidx				
        cell = sheet.cell(self.rowidx, self.colidx)
        self.write_to_cell(sheet_name, self.colidx, self.rowidx, value, data_type)

    def modify_cell_with(self, sheet_name, column, row, op, val):
        """
        Using the sheet name a cell is modified with the given operation and value.
        """
        pass
        # my_sheet_index = self.workbook.sheet_names().index(sheet_name)
        # cell = self.workbook.get_sheet(my_sheet_index).cell(int(row), int(column))
        # curval = cell.value
        # if cell.ctype is XL_CELL_NUMBER:
        #     self.workbook.sheets()
        #     if not self.copied_workbook:
        #         self.copied_workbook = copy(self.workbook)
        #     plain = easyxf('')
        #     modexpr = str(curval) + op + val
        #     self.copied_workbook.get_sheet(my_sheet_index).write(int(row), int(column), eval(modexpr), plain)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rflint.common import KeywordRule, ERROR
from rflint.parser import SettingTable

class RequireKeywordDocumentation(KeywordRule):
    '''Verify that a keyword has documentation'''
    severity=ERROR

    def apply(self, keyword):
        for setting in keyword.settings:
            if setting[1].lower() == "[documentation]" and len(setting) > 2:
                return

        # set the line number to the line immediately after the keyword name
        self.report(keyword, "No keyword documentation", keyword.linenumber+1)


class TooFewKeywordSteps(KeywordRule):
    '''Keywords should have at least a minimum number of steps

    This rule is configurable. The default number of required steps is 2.
    '''

    min_required = 2

    def configure(self, min_required):
        self.min_required = int(min_required)

    def apply(self, keyword):
        # ignore empty steps
        steps = [step for step in keyword.steps if not (len(step) == 1 and not step[0])]
        if len(steps) < self.min_required:
            msg = "Too few steps (%s) in keyword" % len(steps)
            self.report(keyword, msg, keyword.linenumber)




import requests
import json
import re
from robot.api import logger

class Notify:

    ROBOT_LISTENER_API_VERSION = 2
    
    flag = False
    url = "https://slack.com/api/chat.postMessage"
    

    def __init__(self, slackChannel, slackToken, testSuite, buildBy, buildWhy):
        
        self.slackChannel = slackChannel
        self.slackToken = 'Bearer ' + slackToken
        self.testSuite = testSuite
        self.buildBy = buildBy
        self.buildWhy = buildWhy
        self._init_slack_message()
        #self.flag = True


    def _init_slack_message(self):
        self.payload = {}
       

    def end_suite(self, name, result):
        #if self.flag is True:
        #self.suite_name = name
        self.suite_statistics =  result['statistics']
        self.suite_statistics = re.split("\n", self.suite_statistics, 1)
        self.suite_status = result['status']                 
            
    def close(self):
        self._notify_slack()
        
        
    def _notify_slack(self):  
        
        self.headers = {"Authorization":self.slackToken, "Content-Type":"application/json"}   
        
        self.payload['channel'] = self.slackChannel
        
        self.payload['attachments'] =  [{}]
        
        self.payload['attachments'][0]['color'] = '#D3D3D3'
        self.payload['attachments'][0]['blocks'] =  [{},{},{},{},{}]
        
        
        self.payload['attachments'][0]['blocks'][0]['type'] = 'divider'
        self.payload['attachments'][0]['blocks'][1]['type'] = 'context'
        self.payload['attachments'][0]['blocks'][1]['elements'] = [{}]
        self.payload['attachments'][0]['blocks'][1]['elements'][0]['type'] = 'mrkdwn'
        self.payload['attachments'][0]['blocks'][1]['elements'][0]['text'] = '*Test Suite:*'
        
        self.payload['attachments'][0]['blocks'][2]['type'] = 'context'
        self.payload['attachments'][0]['blocks'][2]['elements'] = [{}]
        self.payload['attachments'][0]['blocks'][2]['elements'][0]['type'] = 'mrkdwn'
        self.payload['attachments'][0]['blocks'][2]['elements'][0]['text'] = '*Test Status:*'
        
        
        self.payload['attachments'][0]['blocks'][3]['type'] = 'context'
        self.payload['attachments'][0]['blocks'][3]['elements'] = [{}]
        self.payload['attachments'][0]['blocks'][3]['elements'][0]['type'] = 'mrkdwn'
        self.payload['attachments'][0]['blocks'][3]['elements'][0]['text'] = '*Build Details:*'
        
        
        self.payload['attachments'][0]['blocks'][4]['type'] = 'actions'
        self.payload['attachments'][0]['blocks'][4]['elements'] = [{},{},{}]
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['type'] = 'button' 
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['text'] = {}
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['text']['type'] = 'plain_text'
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['text']['text'] = 'View Metrics'
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['text']['emoji'] = True
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['url'] = 'https://www.google.com/view-metrics' 
        self.payload['attachments'][0]['blocks'][4]['elements'][0]['value'] = 'view-metrics' 
        
        
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['type'] = 'button' 
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['text'] = {}
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['text']['type'] = 'plain_text'
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['text']['text'] = 'View Log'
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['text']['emoji'] = True
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['url'] = 'https://www.google.com/view-log' 
        self.payload['attachments'][0]['blocks'][4]['elements'][1]['value'] = 'view-log' 
        
        
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['type'] = 'button' 
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['text'] = {}
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['text']['type'] = 'plain_text'
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['text']['text'] = 'View Report'
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['text']['emoji'] = True
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['url'] = 'https://www.google.com/view-report' 
        self.payload['attachments'][0]['blocks'][4]['elements'][2]['value'] = 'view-report'
        
        
        testSuite = self.payload["attachments"][0]["blocks"][1]["elements"][0]["text"]
        testSummary = self.payload["attachments"][0]["blocks"][2]["elements"][0]["text"]
        buildDetails = self.payload["attachments"][0]["blocks"][3]["elements"][0]["text"]  
       
        
                
        self.payload["attachments"][0]["blocks"][1]["elements"][0]["text"] = testSuite + " " + self.testSuite
        self.payload["attachments"][0]["blocks"][2]["elements"][0]["text"] = testSummary + " " + self.suite_statistics[1]
        self.payload["attachments"][0]["blocks"][3]["elements"][0]["text"] = buildDetails + " " + "by " + self.buildBy + " - " + self.buildWhy
        
      
        
        if self.suite_status == 'PASS':
            self.payload["attachments"][0]["color"] = "#008000"
        
        else:
            self.payload["attachments"][0]["color"] = "#FF0000"
                 
                     
        response = requests.post(self.url, data=json.dumps(self.payload), headers=self.headers)
        
        if response.status_code == 200:
            logger.info('Slack Notification Passed')
            self.flag = False
            return None
        else:
            logger.error("Slack Notification Failed:", response.status_code, response.reason)
            return None  



# -*- coding: utf-8 -*-

import os
from rbfMobile.keywords import *
from .version import __version__




class rbfMobile(
    _LoggingKeywords,
    _RunOnFailureKeywords,
    _ElementKeywords,
    _ScreenshotKeywords,
    _ApplicationManagementKeywords,
    _WaitingKeywords,
    _TouchKeywords,
    _KeyeventKeywords,
    _AndroidUtilsKeywords,
):
    """rbfMobile is a Mobile App testing library for Robot Framework.

    = Locating or specifying elements =

    All keywords in rbfMobile that need to find an element on the page
    take an argument, either a ``locator`` or a ``webelement``. ``locator``
    is a string that describes how to locate an element using a syntax
    specifying different location strategies. ``webelement`` is a variable that
    holds a WebElement instance, which is a representation of the element.

    == Using locators ==

    By default, when a locator is provided, it is matched against the key attributes
    of the particular element type. For iOS and Android, key attribute is ``id`` for
    all elements and locating elements is easy using just the ``id``. For example:

    | Click Element    id=my_element

    ``id`` and ``xpath`` are not required to be specified,
    however ``xpath`` should start with ``//`` else just use ``xpath`` locator as explained below.

    For example:

    | Click Element    my_element
    | Wait Until Page Contains Element    //*[@type="android.widget.EditText"]


    Appium additionally supports some of the [https://w3c.github.io/webdriver/webdriver-spec.html|Mobile JSON Wire Protocol] locator strategies.
    It is also possible to specify the approach rbfMobile should take
    to find an element by specifying a lookup strategy with a locator
    prefix. Supported strategies are:

    | *Strategy*        | *Example*                                                      | *Description*                     | *Note*                      |
    | identifier        | Click Element `|` identifier=my_element                        | Matches by @id attribute          |                             |
    | id                | Click Element `|` id=my_element                                | Matches by @resource-id attribute |                             |
    | accessibility_id  | Click Element `|` accessibility_id=button3                     | Accessibility options utilize.    |                             |
    | xpath             | Click Element `|` xpath=//UIATableView/UIATableCell/UIAButton  | Matches with arbitrary XPath      |                             |
    | class             | Click Element `|` class=UIAPickerWheel                         | Matches by class                  |                             |
    | android           | Click Element `|` android=UiSelector().description('Apps')     | Matches by Android UI Automator   |                             |
    | ios               | Click Element `|` ios=.buttons().withName('Apps')              | Matches by iOS UI Automation      |                             |
    | nsp               | Click Element `|` nsp=name=="login"                            | Matches by iOSNsPredicate         | Check PR: #196              |
    | css               | Click Element `|` css=.green_button                            | Matches by css in webview         |                             |
    | name              | Click Element `|` name=my_element                              | Matches by @name attribute        | *Only valid* for Selendroid |

    == Using webelements ==

    Starting with version 1.4 of the rbfMobile, one can pass an argument
    that contains a WebElement instead of a string locator. To get a WebElement,
    use the new `Get WebElements` or `Get WebElement` keyword.

    For example:
    | @{elements}    Get Webelements    class=UIAButton
    | Click Element    @{elements}[2]

    """

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = __version__

    def __init__(self, timeout=5, run_on_failure='Capture Page Screenshot'):
        """rbfMobile can be imported with optional arguments.

        ``timeout`` is the default timeout used to wait for all waiting actions.
        It can be later set with `Set Appium Timeout`.

        ``run_on_failure`` specifies the name of a keyword (from any available
        libraries) to execute when a rbfMobile keyword fails.

        By default `Capture Page Screenshot` will be used to take a screenshot of the current page.
        Using the value `No Operation` will disable this feature altogether. See
        `Register Keyword To Run On Failure` keyword for more information about this
        functionality.

        Examples:
        | Library | rbfMobile | 10 | # Sets default timeout to 10 seconds                                                                             |
        | Library | rbfMobile | timeout=10 | run_on_failure=No Operation | # Sets default timeout to 10 seconds and does nothing on failure           |
        """
        for base in rbfMobile.__bases__:
            base.__init__(self)
        self.set_appium_timeout(timeout)
        self.register_keyword_to_run_on_failure(run_on_failure)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings


class Deprecated(object):

    def __init__(self, old_name, new_name):
        self.old_name = old_name
        self.new_name = new_name

    def __get__(self, instance, owner):
        self._warn()
        return getattr(instance, self.new_name)

    def __set__(self, instance, value):
        self._warn()
        setattr(instance, self.new_name, value)

    def _warn(self):
        warnings.warn('"rbfWEB.%s" is deprecated, use '
                      '"rbfWEB.%s" instead.'
                      % (self.old_name, self.new_name),
                      DeprecationWarning)
